# -*- coding: utf-8 -*-
"""
Created on Tue Apr  9 15:48:17 2013

"""
import numpy as np
import scipy.stats as ss
import matplotlib.pyplot as plt
from lhd import lhd

__version_info__ = (0, 9)
__version__ = '.'.join(map(str, __version_info__))

__author__ = 'Abraham Lee'

__all__ = [
    # core functions
    'uv', 'covariance_matrix', 'correlation_matrix',
    # continuous distribution constructors
    'N',
    'U',
    'Exp',
    'Gamma',
    'Beta',
    'LogN',
    'X2',
    'F',
    'Tri',
    'T',
    'Weib',
    # discrete distribution constructors
    'Bern',
    'B',
    'G',
    'H',
    'Pois'
    ]

npts = 10000

CONSTANT_TYPES = (float, int, long)

class NotUpcast(Exception):
    'Raised when an object cannot be converted to a number with uncertainty'

def to_uncertain_func(x):
    """
    Transforms x into an UncertainFunction-compatible object,
    unless it is already an UncertainFunction (in which case x is returned 
    unchanged).

    Raises an exception unless 'x' belongs to some specific classes of
    objects that are known not to depend on UncertainFunction objects
    (which then cannot be considered as constants).
    """
    if isinstance(x, UncertainFunction):
        return x

    #! In Python 2.6+, numbers.Number could be used instead, here:
    elif isinstance(x, CONSTANT_TYPES):
        # No variable => no derivative to define:
        return UncertainFunction(x, [x]*npts)
    
    raise NotUpcast("%s cannot be converted to a number with"
                    " uncertainty" % type(x))
    
    

class UncertainFunction(object):
    """
    UncertainFunction objects represent the uncertainty of a result of 
    calculations with uncertain variables. Nearly all basic mathematical
    operations are supported.
    
    This class is mostly intended for internal use.
    
    
    """
    def __init__(self, x, mcpts):
        self._mcpts = np.atleast_1d(mcpts).flatten()
        self.tag = None

    @property
    def mean(self):
        """
        Mean value as a result of an uncertainty calculation
        """
        mn = np.mean(self._mcpts)
        return mn
    
    @property
    def var(self):
        """
        Variance value as a result of an uncertainty calculation
        """
        mn = self.mean
        vr = np.mean((self._mcpts - mn)**2)
        return vr
        
    @property
    def std(self):
        """
        Standard deviation value as a result of an uncertainty calculation, 
        defined as::
            
                    ________
            std = \/variance
            
        """
        return self.var**0.5
    
    @property
    def skew(self):
        """
        Skewness coefficient value as a result of an uncertainty calculation,
        defined as::
            
              _____     m3
            \/beta1 = ------
                      std**3
        
        where m3 is the third central moment and std is the standard deviation
        """
        mn = self.mean
        sd = self.std
        sk = 0.0 if abs(sd)<=1e-8 else np.mean((self._mcpts - mn)**3)/sd**3
        return sk
    
    @property
    def kurt(self):
        """
        Kurtosis coefficient value as a result of an uncertainty calculation,
        defined as::
            
                      m4
            beta2 = ------
                    std**4

        where m4 is the fourth central moment and std is the standard deviation
        """
        mn = self.mean
        sd = self.std
        kt = 0.0 if abs(sd)<=1e-8 else np.mean((self._mcpts - mn)**4)/sd**4
        return kt
    
    @property
    def stats(self):
        """
        The first four standard moments of a distribution: mean, variance, and
        standardized skewness and kurtosis coefficients.
        """
        mn = self.mean
        vr = self.var
        sk = self.skew
        kt = self.kurt
        return [mn, vr, sk, kt]
        
    def _to_general_representation(self,str_func):
        mn, vr, sk, kt = self.stats
        return ('uv({:}, {:}, {:}, {:})'.format(
            str_func(mn), str_func(vr), str_func(sk), str_func(kt)) 
            if any([vr, sk, kt]) else str_func(mn))

    def __str__(self):
        return self._to_general_representation(str)

    def __repr__(self):
#        return self._to_general_representation(repr)
        return str(self)

    def describe(self, name=None):
        """
        Cleanly show what the four displayed distribution moments are:
            - Mean
            - Variance
            - Standardized Skewness Coefficient
            - Standardized Kurtosis Coefficient
        
        For a standard Normal distribution, these are [0, 1, 0, 3].
        
        If the object has an associated tag, this is presented. If the optional
        ``name`` kwarg is utilized, this is presented as with the moments.
        Otherwise, no unique name is presented.
        
        Example
        =======
        ::
        
            >>> x = N(0, 1, 'x')
            >>> x.describe()  # print tag since assigned
            MCERP Uncertain Value (x):
            ...

            >>> x.describe('foobar')  # 'name' kwarg takes precedence
            MCERP Uncertain Value (foobar):
            ...
            
            >>> y = x**2
            >>> y.describe('y')  # print name since assigned
            MCERP Uncertain Value (y):
            ...

            >>> y.describe()  # print nothing since no tag
            MCERP Uncertain Value:
            ...

         """
        mn, vr, sk, kt = self.stats
        if name is not None:
            s = 'MCERP Uncertain Value ('+name+'):\n'
        elif self.tag is not None:
            s = 'MCERP Uncertain Value ('+self.tag+'):\n'
        else:
            s = 'MCERP Uncertain Value:\n'
        s += ' > Mean................... {: }\n'.format(mn)
        s += ' > Variance............... {: }\n'.format(vr)
        s += ' > Skewness Coefficient... {: }\n'.format(sk)
        s += ' > Kurtosis Coefficient... {: }\n'.format(kt)
        print s
        
    def plot(self, hist=False, **kwargs):
        """
        Plot the distribution of the UncertainFunction. By default, the
        distribution is shown with a kernel density estimate (kde).
        
        Optional
        --------
        hist : bool
            If true, a histogram is displayed
        kwargs : any valid matplotlib.pyplot.plot kwarg
        
        """
        vals = self._mcpts
        low = min(vals)
        high = max(vals)

        p = ss.kde.gaussian_kde(vals)
        xp = np.linspace(low,high,100)

        plt.figure()
        if hist:
            plt.hist(vals, bins=np.round(np.sqrt(vals)), histtype='stepfilled',
                **kwargs)
            if self.tag is not None:
                plt.title('Histogram of ('+self.tag+') - '+str(self))
            else:
                plt.title('Histogram of '+str(self))
        else:
            plt.plot(xp,p.evaluate(xp))
            if self.tag is not None:
                plt.title('KDE of ('+self.tag+') - '+str(self))
            else:
                plt.title('KDE of '+str(self))

        plt.xlim(low - (high - low)*0.1, high + (high - low)*0.1)
        plt.ylim(ymin=0)
        plt.show()

    def __add__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts + uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __radd__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts + uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __mul__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts * uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __rmul__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts * uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __sub__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts - uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __rsub__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[1]._mcpts - uf[0]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __div__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts/uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __rdiv__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[1]._mcpts/uf[0]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __truediv__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts/uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __rtruediv__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[1]._mcpts/uf[0]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __pow__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[0]._mcpts**uf[1]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)

    def __rpow__(self, val):
        uf = map(to_uncertain_func, [self, val])
        mcpts = uf[1]._mcpts**uf[0]._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
    
    def __neg__(self):
        mcpts = -self._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
        
    def __pos__(self):
        mcpts = self._mcpts
        return UncertainFunction(np.mean(mcpts), mcpts)
    
    def __abs__(self):
        mcpts = np.abs(self._mcpts)
        return UncertainFunction(np.mean(mcpts), mcpts)
    
    def __eq__(self,val):
        diff = self - val
        return not (diff.mean or diff.var or diff.skew or diff.kurt)
    
    def __ne__(self,val):
        return not self==val
    
    def __lt__(self,val):
        self,val = map(to_uncertain_func, [self,val])
        return True if float(self.mean - val.mean) < 0 else False
    
    def __le__(self,val):
        return (self==val) or self < val
    
    def __gt__(self,val):
        return not self < val
    
    def __ge__(self,val):
        return (self==val) or self > val

    def __nonzero__(self):
        return self!=0

################################################################################

class UncertainVariable(UncertainFunction):
    """
    UncertainVariable objects track the effects of uncertainty, characterized 
    in terms of the first four standard moments of statistical distributions 
    (mean, variance, skewness and kurtosis coefficients). Monte Carlo simulation,
    in conjunction with Latin-hypercube based sampling performs the calculations.

    Parameters
    ----------
    rv : scipy.stats.distribution
        A distribution to characterize the uncertainty
    
    tag : str, optional
        A string identifier when information about this variable is printed to
        the screen
        
    Notes
    -----
    
    The ``scipy.stats`` module contains many distributions which we can use to
    perform any necessary uncertainty calculation. It is important to follow
    the initialization syntax for creating any kind of distribution object:
        
        - *Location* and *Scale* values must use the kwargs ``loc`` and 
          ``scale``
        - *Shape* values are passed in as non-keyword arguments before the 
          location and scale, (see below for syntax examples)..
        
    The mathematical operations that can be performed on uncertain objects will 
    work for any distribution supplied, but may be misleading if the supplied 
    moments or distribution is not accurately defined. Here are some guidelines 
    for creating UncertainVariable objects using some of the most common 
    statistical distributions:
    
    =========================  ============  ==================  ====  ========
    Distribution               scipy.stats   args                loc   scale
                               class name    (shape parameters)
    =========================  ============  ==================  ====  ========
    Normal(mu, sigma)          norm                              mu    sigma
    Uniform(a, b)              uniform                           a     b-a
    Exponential(lamda)         expon                                   1/lambda
    Gamma(k, theta)            gamma         k                         theta
    Beta(alpha, beta, [a, b])  beta          alpha, beta         a     b-a
    Log-Normal(mu, sigma)      lognorm       sigma               mu
    Chi-Square(k)              chi2          k
    F(d1, d2)                  f             d1, d2
    Triangular(a, b, c)        triang        c                   a     b-a
    Student-T(v)               t             v
    Weibull(lamda, k)          exponweib     lamda, k
    =========================  ============  =================== ====  ========
    Bernoulli(p)               bernoulli     p
    Binomial(n, p)             binomial      n, p
    Geometric(p)               geom          p
    Hypergeometric(M, n, N)    hypergeom     M, n, N
    Poisson(lamda)             poisson       lamda
    =========================  ============  =================== ====  ========
    
    Thus, each distribution above would have the same call signature::
        
        >>> import scipy.stats as ss
        >>> ss.your_dist_here(args, loc=loc, scale=scale)
        
    ANY SCIPY.STATS.DISTRIBUTION SHOULD WORK! IF ONE DOESN'T, PLEASE LET ME
    KNOW!
    
    Convenient constructors have been created to make assigning these 
    distributions easier. They follow the parameter notation found in the
    respective Wikipedia articles:
    
    ===========================  ===========================================
    MCERP Distibution            Wikipedia page
    ===========================  ===========================================
    N(mu, sigma)                 http://en.wikipedia.org/wiki/Normal_distribution
    U(a, b)                      http://en.wikipedia.org/wiki/Uniform_distribution_(continuous)
    Exp(lamda, [mu])             http://en.wikipedia.org/wiki/Exponential_distribution
    Gamma(k, theta)              http://en.wikipedia.org/wiki/Gamma_distribution
    Beta(alpha, beta, [a, b])    http://en.wikipedia.org/wiki/Beta_distribution
    LogN(mu, sigma)              http://en.wikipedia.org/wiki/Log-normal_distribution
    X2(df)                       http://en.wikipedia.org/wiki/Chi-squared_distribution
    F(dfn, dfd)                  http://en.wikipedia.org/wiki/F-distribution
    Tri(a, b, c)                 http://en.wikipedia.org/wiki/Triangular_distribution
    T(df)                        http://en.wikipedia.org/wiki/Student's_t-distribution
    Weib(lamda, k)               http://en.wikipedia.org/wiki/Weibull_distribution
    ===========================  ===========================================
    Bern(p)                      http://en.wikipedia.org/wiki/Bernoulli_distribution
    B(n, p)                      http://en.wikipedia.org/wiki/Binomial_distribution
    G(p)                         http://en.wikipedia.org/wiki/Geometric_distribution
    H(M, n, N)                   http://en.wikipedia.org/wiki/Hypergeometric_distribution
    Pois(lamda)                  http://en.wikipedia.org/wiki/Poisson_distribution
    ===========================  ===========================================

    Examples
    --------
    A three-part assembly
        
        >>> x1 = N(24, 1)
        >>> x2 = N(37, 4)
        >>> x3 = Exp(2)  # Exp(mu=0.5) works too
        
        >>> Z = (x1*x2**2)/(15*(1.5 + x3))
        >>> Z
        uv(1161.46231679, 116646.762981, 0.345533974771, 3.00791101068)

    The result shows the mean, variance, and standardized skewness and kurtosis
    of the output variable Z, which will vary from use to use due to the random
    nature of Monte Carlo simulation and latin-hypercube sampling techniques.
    
    Basic math operations may be applied to distributions, where all 
    statistical calculations are performed using latin-hypercube enhanced Monte
    Carlo simulation. Nearly all of the built-in trigonometric-, logarithm-, 
    etc. functions of the ``math`` module have uncertainty-compatible 
    counterparts that should be used when possible since they support both 
    scalar values and uncertain objects. These can be used after importing the 
    ``umath`` module::
        
        >>> from mcerp.umath import * # sin(), sqrt(), etc.
        >>> sqrt(x1)
        uv(4.89791765647, 0.0104291897681, -0.0614940614672, 3.00264937735)
    
    At any time, the standardized statistics can be retrieved using::
        
        >>> x1.mean
        >>> x1.var  # x1.std (standard deviation) is also available
        >>> x1.skew
        >>> x1.kurt
    
    or all four together with::
    
        >>> x1.stats
    
    By default, the Monte Carlo simulation uses 10000 samples, but this can be
    changed at any time with::
        
        >>> mcerp.npts = number_of_samples
    
    Any value from 1,000 to 1,000,000 is recommended (more samples means more
    accurate, but also means more time required to perform the calculations). 
    Although it can be changed, since variables retain their samples from one
    calculation to the ne        if hist:
            plt.hist(vals, bins=np.round(np.sqrt(vals)), histtype='stepfilled',
                **kwargs)
        else:
xt, this parameter should be changed before any 
    calculations are performed to ensure parameter compatibility (this may 
    change to be more dynamic in the future, but for now this is how it is).
    
    Also, to see the underlying distribution of the variable, and if matplotlib
    is installed, simply call its plot method::
        
        >>> x1.plot()
    
    Optional kwargs can be any valid kwarg used by matplotlib.pyplot.plot
    
    See Also
    --------
    N, U, Exp, Gamma, Beta, LogN, X2, F, Tri, T, Weib, Bern, B, G, H, Pois
        
    """
    
    def __init__(self, rv, tag=None):
        
        assert hasattr(rv, 'dist'), 'Input must be a  distribution from ' + \
            'the scipy.stats module.'
        self.rv = rv
        
        # generate the latin-hypercube points
        self._mcpts = lhd(dist=self.rv, size=npts).flatten()
        self.tag = tag
        
    @property
    def mean(self):
        return float(self.rv.stats('m'))
    
    @property
    def var(self):
        return float(self.rv.stats('v'))
		
    @property
    def std(self):
        return self.var**0.5
        
    @property
    def skew(self):
        return float(self.rv.stats('s'))
    
    @property
    def kurt(self):
        return float(self.rv.stats('k')) + 3  # remove the 3 for standardization
    
    @property
    def stats(self):
        return [self.mean, self.var, self.skew, self.kurt]
    
    def plot(self, hist=False, **kwargs):
        """
        Plot the distribution of the UncertainVariable. Continuous 
        distributions are plotted with a line plot and discrete distributions
        are plotted with discrete circles.
        
        Optional
        --------
        hist : bool
            If true, a histogram is displayed
        kwargs : any valid matplotlib.pyplot.plot kwarg
        
        """
        
        if hist:
            vals = self._mcpts
            plt.hist(vals, bins=np.round(np.sqrt(vals)), histtype='stepfilled',
                **kwargs)

            if self.tag is not None:
                plt.title('Histogram of ('+self.tag+') - '+str(self))
            else:
                plt.title('Histogram of '+str(self))

        else:
            low = int(self.rv.ppf(1e-6))
            high = int(self.rv.ppf(1 - 1e-6))
            if hasattr(self.rv.dist, 'pmf'):
                vals = range(low, high + 1)
                plt.plot(vals, self.rv.pmf(vals), 'o', **kwargs)

                if self.tag is not None:
                    plt.title('PMF of ('+self.tag+') - '+str(self))
                else:
                    plt.title('PMF of '+str(self))

            else:
                vals = np.linspace(low, high, 500)
                plt.plot(vals, self.rv.pdf(vals), **kwargs)

                if self.tag is not None:
                    plt.title('PDF of ('+self.tag+') - '+str(self))
                else:
                    plt.title('PDF of '+str(self))

        plt.xlim(low - (high - low)*0.1, high + (high - low)*0.1)
        plt.show()
                
        
uv = UncertainVariable # a nicer form for the user

###############################################################################
# Define some convenience constructors for common statistical distributions.
# Hopefully these are a little easier/more intuitive to use than the 
# scipy.stats.distributions.
###############################################################################

def N(mu, sigma, tag=None):
    """
    A normal random variate
    
    Parameters
    ----------
    mu : scalar
        The mean value of the distribution
    sigma : scalar
        The standard deviation (must be positive and non-zero)
    """
    assert sigma>0, 'Sigma must be positive'
    return uv(ss.norm(loc=mu, scale=sigma), tag=tag)

###############################################################################

def U(a, b, tag=None):
    """
    A uniform random variate
    
    Parameters
    ----------
    low : scalar
        Lower bound of the distribution support.
    high : scalar
        Upper bound of the distribution support.
    """
    assert a<b, 'Lower bound must be less than the upper bound'
    return uv(ss.uniform(loc=a, scale=b-a), tag=tag)

###############################################################################

def Exp(lamda, mu=None, tag=None):
    """
    An exponential random variate
    
    Parameters
    ----------
    lamda : scalar
        The inverse scale (as shown on Wikipedia). 
    
    Optional
    --------
    mu : scalar
        The mean value of the distribution (must be positive and non-zero). If 
        this is given, ``lamda`` is ignored. (FYI: mu = 1/lamda.)
    """
    if mu is not None:
        assert mu>0, 'Mean must be positive and not zero'
        return uv(ss.expon(scale=mu), tag=tag)
    else:
        return uv(ss.expon(scale=1./lamda), tag=tag)

###############################################################################

def Gamma(k, theta, tag=None):
    """
    A gamma random variate
    
    Parameters
    ----------
    k : scalar
        The shape parameter (must be positive and non-zero)
    theta : scalar
        The scale parameter (must be positive and non-zero)
    """
    assert k>0 and theta>0, 'Gamma parameters must be greater than zero'
    return uv(ss.gamma(k, scale=theta), tag=tag)

###############################################################################

def Beta(alpha, beta, a=0, b=1, tag=None):
    """
    A beta random variate
    
    Parameters
    ----------
    alpha : scalar
        The first shape parameter
    beta : scalar
        The second shape parameter
    
    Optional
    --------
    a : scalar
        Lower bound of the distribution support (default=0)
    b : scalar
        Upper bound of the distribution support (default=1)
    """
    assert alpha>0 and beta>0, 'Shape parameters must be greater than zero'
    return uv(ss.beta(alpha, beta, loc=a, scale=b-a), tag=tag)

###############################################################################

def LogN(mu, sigma, tag=None):
    """
    A log-normal random variate
    
    Parameters
    ----------
    mu : scalar
        The location parameter
    sigma : scalar
        The scale parameter (must be positive and non-zero)
    """
    assert sigma>0, 'Sigma must be positive'
    return uv(ss.lognorm(sigma, loc=mu), tag=tag)

###############################################################################

def X2(k, tag=None):
    """
    An Chi-Squared random variate
    
    Parameters
    ----------
    k : int
        The degrees of freedom of the distribution (must be greater than one)
    """
    assert isinstance(df, int) and df>1, 'DF must be an int greater than 1'
    return uv(ss.chi2(df), tag=tag)

###############################################################################

def F(d1, d2, tag=None):
    """
    An F (fisher) random variate
    
    Parameters
    ----------
    d1 : int
        Numerator degrees of freedom
    d2 : int
        Denominator degrees of freedom
    """
    assert isinstance(d1, int) and d1>1, 'DFN must be an int greater than 1'
    assert isinstance(d2, int) and d2>1, 'DFD must be an int greater than 1'
    return uv(ss.f(d1, d2), tag=tag)

###############################################################################

def Tri(a, b, c, tag=None):
    """
    A triangular random variate
    
    Parameters
    ----------
    a : scalar
        Lower bound of the distribution support (default=0)
    b : scalar
        Upper bound of the distribution support (default=1)
    c : scalar
        The location of the triangle's peak (a <= c <= b)
    """
    assert a<=c<=b, 'peak must lie in between low and high'
    return uv(ss.triang(c, loc=a, scale=b-a), tag=tag)

###############################################################################

def T(v, tag=None):
    """
    An Student-T random variate
    
    Parameters
    ----------
    v : int
        The degrees of freedom of the distribution (must be greater than one)
    """
    assert isinstance(v, int) and v>1, 'DF must be an int greater than 1'
    return uv(ss.t(v), tag=tag)

###############################################################################

def Weib(lamda, k, tag=None):
    """
    An Weibull random variate
    
    Parameters
    ----------
    lamda : scalar
        The scale parameter
    k : scalar
        The shape parameter
    """
    assert lamda>0 and k>0, 'Weibull scale and shape parameters must be greater than zero'
    return uv(ss.exponweib(lamda, k), tag=tag)

###############################################################################

def Bern(p, tag=None):
    """
    An Bernoulli random variate
    
    Parameters
    ----------
    p : scalar
        The probability of success
    """
    assert 0<p<1, 'Bernoulli probability must be between zero and one'
    return uv(ss.bernoulli(p), tag=tag)

###############################################################################

def B(n, p, tag=None):
    """
    An Binomial random variate
    
    Parameters
    ----------
    n : int
        The number of trials
    p : scalar
        The probability of success
    """
    assert int(n)==n and n>0, 'Binomial number of trials must be an integer greater than zero'
    assert 0<p<1, 'Binomial probability must be between zero and one'
    return uv(ss.binom(n, p), tag=tag)

###############################################################################

def G(p, tag=None):
    """
    An Geometric random variate
    
    Parameters
    ----------
    p : scalar
        The probability of success
    """
    assert 0<p<1, 'Geometric probability must be between zero and one'
    return uv(ss.geom(p), tag=tag)

###############################################################################

def H(p, tag=None):
    """
    An Hypergeometric random variate
    
    Parameters
    ----------
    M : int
        The total population size
    n : int
        The number of individuals of interest in the population
    N : int
        The number of individuals that will be chosen from the population
    """
    assert int(M)==M and M>0, 'Hypergeometric total population size must be an integer greater than zero.'
    assert int(n)==n and 0<n<=M, 'Hypergeometric interest population size must be an integer greater than zero and no more than the total population size.'
    assert int(N)==N and 0<N<=M, 'Hypergeometric chosen population size must be an integer greater than zero and no more than the total population size.'
    return uv(ss.hypergeom(M, n, N), tag=tag)

###############################################################################

def Pois(lamda, tag=None):
    """
    An Poisson random variate
    
    Parameters
    ----------
    lamda : scalar
        The rate of an occurance within a specified interval of time or space.
    """
    assert lamda>0, 'Poisson rate must be greater than zero.'
    return uv(ss.poisson(lamda), tag=tag)

###############################################################################

def covariance_matrix(nums_with_uncert):
    """
    Calculate the covariance matrix of uncertain variables, oriented by the
    order of the inputs
    
    Parameters
    ----------
    nums_with_uncert : array-like
        A list of variables that have an associated uncertainty
    
    Returns
    -------
    cov_matrix : 2d-array-like
        A nested list containing covariance values
    
    Example
    -------
    
        >>> x = N(1, 0.1)
        >>> y = N(10, 0.1)
        >>> z = x + 2*y
        >>> covariance_matrix([x,y,z])
        [[  9.99694861e-03   2.54000840e-05   1.00477488e-02]
         [  2.54000840e-05   9.99823207e-03   2.00218642e-02]
         [  1.00477488e-02   2.00218642e-02   5.00914772e-02]]

    """
    ufuncs = map(to_uncertain_func,nums_with_uncert)
    cov_matrix = []
    for (i1, expr1) in enumerate(ufuncs):
        coefs_expr1 = []
        mean1 = expr1.mean
        for (i2, expr2) in enumerate(ufuncs[:i1+1]):
            mean2 = expr2.mean
            coef = np.mean((expr1._mcpts - mean1)*(expr2._mcpts - mean2))
            coefs_expr1.append(coef)
        cov_matrix.append(coefs_expr1)
        
    # We symmetrize the matrix:
    for (i, covariance_coefs) in enumerate(cov_matrix):
        covariance_coefs.extend(cov_matrix[j][i]
                                for j in range(i+1, len(cov_matrix)))

    return cov_matrix

def correlation_matrix(nums_with_uncert):
    """
    Calculate the correlation matrix of uncertain variables, oriented by the
    order of the inputs
    
    Parameters
    ----------
    nums_with_uncert : array-like
        A list of variables that have an associated uncertainty
    
    Returns
    -------
    corr_matrix : 2d-array-like
        A nested list containing covariance values
    
    Example
    -------
    
        >>> x = N(1, 0.1)
        >>> y = N(10, 0.1)
        >>> z = x + 2*y
        >>> correlation_matrix([x,y,z])
        [[ 0.99969486  0.00254001  0.4489385 ]
         [ 0.00254001  0.99982321  0.89458702]
         [ 0.4489385   0.89458702  1.        ]]

    """
    ufuncs = map(to_uncertain_func,nums_with_uncert)
    cov_matrix = covariance_matrix(ufuncs)
    corr_matrix = []
    for (i1, expr1) in enumerate(ufuncs):
        row_data = []
        for (i2, expr2) in enumerate(ufuncs):
            row_data.append(cov_matrix[i1][i2]/expr1.std/expr2.std)
        corr_matrix.append(row_data)
    return corr_matrix
    
if __name__=='__main__':
    
    import mcerp.umath as umath
    
    print '*'*80
    print 'TEST FUNCTIONS USING DERIVED MOMENTS FROM SCIPY DISTRIBUTIONS'
    print '*'*80
    print 'Example of a three part assembly'
    x1 = N(24, 1)
    x2 = N(37, 4)
    x3 = Exp(2)  # Exp(mu=0.5) is the same
    Z = (x1*x2**2)/(15*(1.5 + x3))
    Z.describe()
    
    print '*'*80
    print 'Example of volumetric gas flow through orifice meter'
    H = N(64, 0.5)
    M = N(16, 0.1)
    P = N(361, 2)
    t = N(165, 0.5)
    C = 38.4
    Q = C*umath.sqrt((520*H*P)/(M*(t + 460)))
    Q.describe()

    print '*'*80
    print 'Example of manufacturing tolerance stackup'
    # for a gamma distribution we need the following conversions:
    # shape = mean**2/var
    # scale = var/mean
    mn = 1.5
    vr = 0.25
    k = mn**2/vr
    theta = vr/mn
    x = Gamma(k, theta)
    y = Gamma(k, theta)
    z = Gamma(k, theta)
    w = x + y + z
    w.describe()

    print '*'*80
    print 'Example of scheduling facilities (six stations)'
    s1 = N(10, 1)
    s2 = N(20, 2**0.5)
    mn1 = 1.5
    vr1 = 0.25
    k1 = mn1**2/vr1
    theta1 = vr1/mn1
    s3 = Gamma(k1, theta1)
    mn2 = 10
    vr2 = 10
    k2 = mn2**2/vr2
    theta2 = vr2/mn2
    s4 = Gamma(k2, theta2)
    s5 = Exp(5)  # Exp(mu=0.2) is the same
    s6 = X2(10)
    T = s1 + s2 + s3 + s4 + s5 + s6
    T.describe()

    print '*'*80
    print 'Example of two-bar truss stress/deflection analysis'
    H = N(30, 5/3., tag='H')
    B = N(60, 0.5/3., tag='B')
    d = N(3, 0.1/3, tag='d')
    t = N(0.15, 0.01/3, tag='t')
    E = N(30000, 1500/3., tag='E')
    rho = N(0.3, 0.01/3., tag='rho')
    P = N(66, 3/3., tag='P')
    pi = np.pi
    wght = 2*pi*rho*d*t*umath.sqrt((B/2)**2 + H**2)
    strs = (P*umath.sqrt((B/2)**2 + H**2))/(2*pi*d*t*H)
    buck = (pi**2*E*(d**2 + t**2))/(8*((B/2)**2 + H**2))
    defl = (P*((B/2)**2 + H**2)**(1.5))/(2*pi*d*t*H**2*E)
    print wght.describe('wght')
    print strs.describe('strs')
    print buck.describe('buck')
    print defl.describe('defl')

    print '** TESTS COMPLETE **'



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 24 05:32:45 2013

@author: tisimst
"""
from copy import copy
import numpy as np
import scipy.stats as ss

def lhd(dist=None,size=None,dims=1,form='randomized',iterations=100,
        showcorrelations=False):
    """
    Create a Latin-Hypercube sample design based on distributions defined in the
    `scipy.stats` module
    
    Parameters
    ----------
    dist: array_like
        frozen scipy.stats.rv_continuous or rv_discrete distribution objects 
        that are defined previous to calling LHD

    size: int
        integer value for the number of samples to generate for each 
        distribution object
        
    dims: int, optional
        if dist is a single distribution object, and dims > 1, the one
        distribution will be used to generate a size-by-dims sampled design
        
    form: str, optional (non-functional at the moment)
        determines how the sampling is to occur, with the following optional 
        values:
            - 'randomized' - completely randomized sampling
            - 'spacefilling' - space-filling sampling (generally gives a more 
              accurate sampling of the design when the number of sample points 
              is small)
            - 'orthogonal' - balanced space-filling sampling (experimental)
              
        The 'spacefilling' and 'orthogonal' forms require some iterations to 
        determine the optimal sampling pattern. 
        
    iterations: int, optional (non-functional at the moment)
        used to control the number of allowable search iterations for generating
        'spacefilling' and 'orthogonal' designs
    
    Returns
    -------
    out: 2d-array, 
        A 2d-array where each column corresponds to each input distribution and 
        each row is a sample in the design
    
    Examples
    --------
    
    Single distribution: 
        - uniform distribution, low = -1, width = 2
        
    >>> import scipy.stats as ss
    >>> d0 = ss.uniform(loc=-1,scale=2)
    >>> print lhd(dist=d0,size=5)
    [[ 0.51031081]
     [-0.28961427]
     [-0.68342107]
     [ 0.69784371]
     [ 0.12248842]]

    Single distribution for multiple variables:
        - normal distribution, mean = 0, stdev = 1
        
    >>> d1 = ss.norm(loc=0,scale=1)
    >>> print lhd(dist=d1,size=7,dims=5)
    [[-0.8612785   0.23034412  0.21808001]
     [ 0.0455778   0.07001606  0.31586419]
     [-0.978553    0.30394663  0.78483995]
     [-0.26415983  0.15235896  0.51462024]
     [ 0.80805686  0.38891031  0.02076505]
     [ 1.63028931  0.52104917  1.48016008]] 
    
    Multiple distributions:
        - beta distribution, alpha = 2, beta = 5
        - exponential distribution, lambda = 1.5
        
    >>> d2 = ss.beta(2,5)
    >>> d3 = ss.expon(scale=1/1.5)
    >>> print lhd(dist=(d1,d2,d3),size=6)
    [[-0.8612785   0.23034412  0.21808001]
     [ 0.0455778   0.07001606  0.31586419]
     [-0.978553    0.30394663  0.78483995]
     [-0.26415983  0.15235896  0.51462024]
     [ 0.80805686  0.38891031  0.02076505]
     [ 1.63028931  0.52104917  1.48016008]]
    """
    assert dims>0,'kwarg "dims" must be at least 1'
    if not size or not dist:
        return None
    
    def _lhs(x,samples=20):
        """
        _lhs(x) returns a latin-hypercube matrix (each row is a different
        set of sample inputs) using a default sample size of 20 for each column
        of X. X must be a 2xN matrix that contains the lower and upper bounds of
        each column. The lower bound(s) should be in the first row and the upper
        bound(s) should be in the second row.
        
        _lhs(x,samples=N) uses the sample size of N instead of the default (20).
        
        Example:
            >>> x = np.array([[0,-1,3],[1,2,6]])
            >>> print 'x:'; print x
            x:
            [[ 0 -1  3]
             [ 1  2  6]]

            >>> print 'lhs(x):'; print _lhs(x)
            lhs(x):
            [[ 0.02989122 -0.93918734  3.14432618]
             [ 0.08869833 -0.82140706  3.19875152]
             [ 0.10627442 -0.66999234  3.33814979]
             [ 0.15202861 -0.44157763  3.57036894]
             [ 0.2067089  -0.34845384  3.66930908]
             [ 0.26542056 -0.23706445  3.76361414]
             [ 0.34201421 -0.00779306  3.90818257]
             [ 0.37891646  0.15458423  4.15031708]
             [ 0.43501575  0.23561118  4.20320064]
             [ 0.4865449   0.36350601  4.45792314]
             [ 0.54804367  0.56069855  4.60911539]
             [ 0.59400712  0.7468415   4.69923486]
             [ 0.63708876  0.9159176   4.83611204]
             [ 0.68819855  0.98596354  4.97659182]
             [ 0.7368695   1.18923511  5.11135111]
             [ 0.78885724  1.28369441  5.2900157 ]
             [ 0.80966513  1.47415703  5.4081971 ]
             [ 0.86196731  1.57844205  5.61067689]
             [ 0.94784517  1.71823504  5.78021164]
             [ 0.96739728  1.94169017  5.88604772]]

            >>> print 'lhs(x,samples=5):'; print _lhs(x,samples=5)
            lhs(x,samples=5):
            [[ 0.1949127  -0.54124725  3.49238369]
             [ 0.21128576 -0.13439798  3.65652016]
             [ 0.47516308  0.39957406  4.5797308 ]
             [ 0.64400392  0.90890999  4.92379431]
             [ 0.96279472  1.79415307  5.52028238]]	
      """
    
    	# determine the segment size
    	segmentSize = 1.0/samples
    
    	# get the number of dimensions to sample (number of columns)
    	numVars = x.shape[1]
    
    	# populate each dimension
    	out = np.zeros((samples,numVars))
    	pointValue = np.zeros(samples)
    
    	for n in range(numVars):
    		for i in range(samples):
    			segmentMin = i*segmentSize
    			point = segmentMin + (np.random.random()*segmentSize)
    			pointValue[i] = (point*(x[1,n]-x[0,n])) + x[0,n]
    		out[:,n] = pointValue
    
    	# now randomly arrange the different segments
    	return _mix(out)
    
    def _mix(data,dim='rows'):
    	"""
    	Takes a data matrix and mixes up the values along dim (either "rows" or "cols")
    	"""
    	tmpdata = copy(data)
    	if dim is 'cols':
    		tmpdata = tmpdata.T
    
    	for k in range(data.shape[1]):
    		for i in range(data.shape[0]):
    			j = np.random.randint(data.shape[0])
    			temp = copy(tmpdata[i,k])
    			tmpdata[i,k] = copy(tmpdata[j,k])
    			tmpdata[j,k] = copy(temp)
    	
    	return tmpdata
    
    if form is 'randomized':
        if hasattr(dist,'__getitem__'): # if multiple distributions were input
            nvars = len(dist)
            x = np.vstack((np.zeros(nvars),np.ones(nvars)))
            unif_data = _lhs(x,samples=size)
            dist_data = np.empty_like(unif_data)
            for i,d in enumerate(dist):
                dist_data[:,i] = d.ppf(unif_data[:,i])
                
        else: # if a single distribution was input
            nvars = dims
            x = np.vstack((np.zeros(nvars),np.ones(nvars)))
            unif_data = _lhs(x,samples=size)
            dist_data = np.empty_like(unif_data)
            for i in range(nvars):
                dist_data[:,i] = dist.ppf(unif_data[:,i])
        
    elif form is 'spacefilling':
        def euclid_distance(arr):
            n = arr.shape[0]
            ans = 0.0
            for i in range(n-1):
                for j in range(i+1,n):
                    d = np.sqrt(np.sum([(arr[i,k]-arr[j,k])**2 for k in range(arr.shape[1])]))
                    ans += 1.0/d**2
            return ans
        
        def fill_space(data):
            best = 1e8
            for it in range(iterations):
                d = euclid_distance(data)
                if d<best:
                    d_opt = d
                    data_opt = data.copy()
                
                data = _mix(data)
            
            print 'Optimized Distance:',d_opt
            return data_opt

        if hasattr(dist,'__getitem__'): # if multiple distributions were input
            nvars = len(dist)
            x = np.vstack((np.zeros(nvars),np.ones(nvars)))
            unif_data = fill_space(_lhs(x,samples=size))
            dist_data = np.empty_like(unif_data)
            for i,d in enumerate(dist):
                dist_data[:,i] = d.ppf(unif_data[:,i])
                
        else: # if a single distribution was input
            nvars = dims
            x = np.vstack((np.zeros(nvars),np.ones(nvars)))
            unif_data = fill_space(_lhs(x,samples=size))
            dist_data = np.empty_like(unif_data)
            for i in range(nvars):
                dist_data[:,i] = dist.ppf(unif_data[:,i])
                

    elif form is 'orthogonal':
        raise NotImplementedError("Sorry. The orthogonal space-filling algorithm hasn't been implemented yet.")
    else:
        raise ValueError('Invalid "form" value: %s'%(form))

    if dist_data.shape[1]>1:
        cor_matrix = np.zeros((nvars,nvars))
        for i in range(nvars):
            for j in range(nvars):
                x_data = dist_data[:,i].copy()
                y_data = dist_data[:,j].copy()
                x_mean = x_data.mean()
                y_mean = y_data.mean()
                num = np.sum((x_data-x_mean)*(y_data-y_mean))
                den = np.sqrt(np.sum((x_data-x_mean)**2)*np.sum((y_data-y_mean)**2))
                cor_matrix[i,j] = num/den
                cor_matrix[j,i] = num/den
        inv_cor_matrix = np.linalg.pinv(cor_matrix)
        VIF = np.max(np.diag(inv_cor_matrix))
            
        if showcorrelations:
            print 'Correlation Matrix:\n',cor_matrix
            print 'Inverted Correlation Matrix:\n',inv_cor_matrix
            print 'Variance Inflation Factor (VIF):',VIF
#        elif VIF >= 1:
#            print 'WARNING: Variance Inflation Factor (%5.3f) indicates that there'%(VIF)
#            print 'may be some undesirably large pairwise correlations present.'
                
    return dist_data
    
if __name__=='__main__':
    # test single distribution
    d0 = ss.uniform(loc=-1,scale=2) # uniform distribution,low=-1, width=2
    print lhd(dist=d0,size=5)
    
    # test single distribution for multiple variables
    d1 = ss.norm(loc=0,scale=1) # normal distribution, mean=0, stdev=1
    print lhd(dist=d1,size=7,dims=5)
    
    # test multiple distributions
    d2 = ss.beta(2,5) # beta distribution, alpha=2, beta=5
    d3 = ss.expon(scale=1/1.5) # exponential distribution, lambda=1.5
    print lhd(dist=(d1,d2,d3),size=6)
    
    rand_lhs = lhd(dist=(d0,d1,d2,d3),size=100)
    spac_lhs = lhd(dist=(d0,d1,d2,d3),size=100,form='spacefilling',
                   iterations=100,showcorrelations=True)
    
    try:
        from scatterplot_matrix import scatterplot_matrix as spm
        import matplotlib.pyplot as plt
    except ImportError:
        print rand_lhs
        print spac_lhs
    else:
        names = ['U(-1,1)','N(0,1)','Beta(2,5)','Exp(1.5)']
        spm(rand_lhs.T,names=names)
        plt.suptitle('Completely Random LHS Design')
        plt.show()
        spm(spac_lhs.T,names=names)
        plt.suptitle('Space-Filling LHS Design')
        plt.show()
        


from distutils.core import setup

with open('README.txt') as file:
    long_description = file.read()

setup(name='mcerp',
    version='0.9',
    author='Abraham Lee',
    description='Real-time Monte Carlo, Latin-Hypercube Sampling-based, Error Propagation',
    author_email='tisimst@gmail.com',
    url='http://pypi.python.org/pypi/mcerp',
    license='BSD License',
    long_description=long_description,
    package_dir={'mcerp':''},
    packages=['mcerp'],
    requires=['numpy','scipy'],
    include_package_data = True,
    package_data = {
        '': ['revision_history.txt'],
        },
    keywords=[
        'monte carlo', 
        'latin hypercube', 
        'sampling calculator', 
        'error propagation', 
        'uncertainty', 
        'error', 
        'real-time'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: OS Independent',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
        ]
    )
